@@ -34,13 +34,13 @@
 
 struct _VinagreVncTabPrivate
 {
-  GtkWidget  *vnc;
+  GtkWidget  *vnc, *align;
   gboolean   pointer_grab;
   gchar      *clipboard_str;
   GSList     *connected_actions, *initialized_actions;
   GtkWidget  *viewonly_button, *scaling_button;
-  GtkAction  *scaling_action, *viewonly_action, *original_size_action;
-  gulong     signal_clipboard;
+  GtkAction  *scaling_action, *viewonly_action, *original_size_action, *keep_ratio_action;
+  gulong     signal_clipboard, signal_align;
 };
 
 G_DEFINE_TYPE (VinagreVncTab, vinagre_vnc_tab, VINAGRE_TYPE_TAB)
@@ -85,6 +85,13 @@ view_scaling_cb (GtkAction *action, VinagreVncTab *vnc_tab)
 			       gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
 }
 
+static void
+view_keep_ratio_cb (GtkAction *action, VinagreVncTab *vnc_tab)
+{
+  vinagre_vnc_tab_set_keep_ratio (vnc_tab,
+				  gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
+}
+
 static void
 view_viewonly_cb (GtkAction *action, VinagreVncTab *vnc_tab)
 {
@@ -259,6 +266,7 @@ open_vnc (VinagreVncTab *vnc_tab)
   GError     *error;
   VncDisplay *vnc = VNC_DISPLAY (vnc_tab->priv->vnc);
   VinagreTab *tab = VINAGRE_TAB (vnc_tab);
+  GtkWindow  *window = GTK_WINDOW (vinagre_tab_get_window (tab));
 
   success = TRUE;
   error = NULL;
@@ -290,12 +298,12 @@ open_vnc (VinagreVncTab *vnc_tab)
   else
     {
       if (ssh_tunnel_host && *ssh_tunnel_host)
-	if (!vinagre_vnc_tunnel_create (&host, &port_str, ssh_tunnel_host, &error))
+	if (!vinagre_vnc_tunnel_create (window, &host, &port_str, ssh_tunnel_host, &error))
 	  {
 	    success = FALSE;
 	    vinagre_utils_show_error (_("Error creating the SSH tunnel"),
 				      error ? error->message : _("Unknown reason"),
-				      GTK_WINDOW (vinagre_tab_get_window (tab)));
+				      window);
 	    goto out;
 	  }
       success = vnc_display_open_host (vnc, host, port_str);
@@ -306,7 +314,7 @@ open_vnc (VinagreVncTab *vnc_tab)
   else
     vinagre_utils_show_error (_("Error connecting to host."),
 			      error ? error->message : _("Unknown reason"),
-			      GTK_WINDOW (vinagre_tab_get_window (tab)));
+			      window);
 
 out:
   g_free (port_str);
@@ -348,7 +356,7 @@ vnc_auth_unsupported_cb (VncDisplay *vnc, guint auth_type, VinagreVncTab *vnc_ta
   name = vinagre_connection_get_best_name (vinagre_tab_get_conn (tab));
   emphasis = g_strdup_printf ("<i>%s</i>", name);
 
-  /* Translators: %s is a host name or IP address; %u is a code error (number). */
+  /* Translators: %s is a host name or IP address; %u is a code error (number). Also, the right sentence would be ... method for host... */
   g_string_printf (message, _("Authentication method to host %s is unsupported. (%u)"),
 		   emphasis,
 		   auth_type);
@@ -409,17 +417,19 @@ vnc_initialized_cb (VncDisplay *vnc, VinagreVncTab *vnc_tab)
 {
   GtkLabel *label;
   gchar    *name;
-  gboolean scaling, view_only, fullscreen;
+  gboolean scaling, view_only, fullscreen, keep_ratio;
   VinagreTab *tab = VINAGRE_TAB (vnc_tab);
   VinagreConnection *conn = vinagre_tab_get_conn (tab);
 
   g_object_get (conn,
 		"view-only", &view_only,
 		"scaling", &scaling,
+		"keep_ratio", &keep_ratio,
 		"fullscreen", &fullscreen,
 		NULL);
 
   gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (vnc_tab->priv->scaling_action), scaling);
+  gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (vnc_tab->priv->keep_ratio_action), keep_ratio);
   gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (vnc_tab->priv->viewonly_action), view_only);
   vnc_display_set_pointer_local (vnc, TRUE);
   vnc_display_set_keyboard_grab (vnc, TRUE);
@@ -441,128 +451,25 @@ vnc_initialized_cb (VncDisplay *vnc, VinagreVncTab *vnc_tab)
   g_signal_emit_by_name (G_OBJECT (tab), "tab-initialized");
 }
 
-typedef struct {
-  GtkWidget *uname, *pw, *button;
-} ControlOKButton;
-
-static void
-control_ok_button (GtkEditable *entry, ControlOKButton *data)
-{
-  gboolean enabled = TRUE;
-
-  if (GTK_WIDGET_VISIBLE (data->uname))
-    enabled = enabled && gtk_entry_get_text_length (GTK_ENTRY (data->uname)) > 0;
-
-  if (GTK_WIDGET_VISIBLE (data->pw))
-    enabled = enabled && gtk_entry_get_text_length (GTK_ENTRY (data->pw)) > 0;
-
-  gtk_widget_set_sensitive (data->button, enabled);
-}
-
-static gboolean
-ask_credential (VinagreVncTab *vnc_tab,
-		gboolean      need_username,
-		gboolean      need_password,
-		gchar         **username,
-		gchar         **password)
-{
-  GtkBuilder      *xml;
-  GtkWidget       *password_dialog, *host_label, *save_credential_check;
-  GtkWidget       *password_label, *username_label, *image;
-  gchar           *name;
-  int             result;
-  ControlOKButton control;
-  VinagreTab      *tab = VINAGRE_TAB (vnc_tab);
-  VinagreConnection *conn = vinagre_tab_get_conn (tab);
-
-  xml = vinagre_utils_get_builder (NULL, NULL);
-
-  password_dialog = GTK_WIDGET (gtk_builder_get_object (xml, "auth_required_dialog"));
-  gtk_window_set_transient_for (GTK_WINDOW(password_dialog),
-				GTK_WINDOW(vinagre_tab_get_window (tab)));
-
-  host_label = GTK_WIDGET (gtk_builder_get_object (xml, "host_label"));
-  name = vinagre_connection_get_best_name (conn);
-  gtk_label_set_label (GTK_LABEL (host_label), name);
-  g_free (name);
-
-  control.uname  = GTK_WIDGET (gtk_builder_get_object (xml, "username_entry"));
-  control.pw     = GTK_WIDGET (gtk_builder_get_object (xml, "password_entry"));
-  control.button = GTK_WIDGET (gtk_builder_get_object (xml, "ok_button"));
-  password_label = GTK_WIDGET (gtk_builder_get_object (xml, "password_label"));
-  username_label = GTK_WIDGET (gtk_builder_get_object (xml, "username_label"));
-  save_credential_check = GTK_WIDGET (gtk_builder_get_object (xml, "save_credential_check"));
-
-  image = gtk_image_new_from_stock (GTK_STOCK_DIALOG_AUTHENTICATION, GTK_ICON_SIZE_BUTTON);
-  gtk_button_set_image (GTK_BUTTON (control.button), image);
-
-  g_signal_connect (control.uname, "changed", G_CALLBACK (control_ok_button), &control);
-  g_signal_connect (control.pw, "changed", G_CALLBACK (control_ok_button), &control);
-
-  if (need_username)
-    {
-      if (*username)
-        gtk_entry_set_text (GTK_ENTRY (control.uname), *username);
-    }
-  else
-    {
-      gtk_widget_hide (username_label);
-      gtk_widget_hide (control.uname);
-    }
-
-  if (need_password)
-    {
-      if (*password)
-        gtk_entry_set_text (GTK_ENTRY (control.pw), *password);
-    }
-  else
-    {
-      gtk_widget_hide (password_label);
-      gtk_widget_hide (control.pw);
-    }
-
-  result = gtk_dialog_run (GTK_DIALOG (password_dialog));
-  if (result == -5)
-    {
-      g_free (*username);
-      if (gtk_entry_get_text_length (GTK_ENTRY (control.uname)) > 0)
-	*username = g_strdup (gtk_entry_get_text (GTK_ENTRY (control.uname)));
-      else
-	*username = NULL;
-
-      g_free (*password);
-      if (gtk_entry_get_text_length (GTK_ENTRY (control.pw)) > 0)
-	*password = g_strdup (gtk_entry_get_text (GTK_ENTRY (control.pw)));
-      else
-	*password = NULL;
-
-      vinagre_tab_set_save_credentials (tab,
-					gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (save_credential_check)));
-    }
-
-  gtk_widget_destroy (GTK_WIDGET (password_dialog));
-  g_object_unref (xml);
-
-  return result == -5;
-}
-
 static void
 vnc_authentication_cb (VncDisplay *vnc, GValueArray *credList, VinagreVncTab *vnc_tab)
 {
-  gchar *username, *password;
-  gboolean need_password, need_username;
+  gchar *username, *password, *host;
+  gboolean need_password, need_username, save_in_keyring;
   int i;
   VinagreTab *tab = VINAGRE_TAB (vnc_tab);
   VinagreConnection *conn = vinagre_tab_get_conn (tab);
-  VinagreWindow *window = vinagre_tab_get_window (tab);
+  GtkWindow *window = GTK_WINDOW (vinagre_tab_get_window (tab));
 
   if (credList == NULL)
     return;
 
   need_password = FALSE;
   need_username = FALSE;
+  save_in_keyring = FALSE;
   username = NULL;
   password = NULL;
+  host = NULL;
 
   for (i = 0; i < credList->n_values; i++) {
     switch (g_value_get_enum (&credList->values[i]))
@@ -596,7 +503,15 @@ vnc_authentication_cb (VncDisplay *vnc, GValueArray *credList, VinagreVncTab *vn
       vinagre_tab_find_credentials_in_keyring (tab, &username, &password);
       if ( (need_username && !username) || (need_password && !password) )
 	{
-	  if (!ask_credential (vnc_tab, need_username, need_password, &username, &password))
+	  host = vinagre_connection_get_best_name (conn);
+	  if (!vinagre_utils_ask_credential (window,
+					     "VNC",
+					     host,
+					     need_username,
+					     need_password,
+					     &username,
+					     &password,
+					     &save_in_keyring))
 	    {
 	      vinagre_tab_remove_from_notebook (tab);
 	      goto out;
@@ -615,7 +530,7 @@ vnc_authentication_cb (VncDisplay *vnc, GValueArray *credList, VinagreVncTab *vn
 	      vinagre_tab_remove_from_notebook (tab);
 	      vinagre_utils_show_error (_("Authentication error"),
 					_("A username is required in order to access this machine."),
-					GTK_WINDOW (window));
+					window);
 	      goto out;
 	    }
 	}
@@ -632,14 +547,17 @@ vnc_authentication_cb (VncDisplay *vnc, GValueArray *credList, VinagreVncTab *vn
 	      vinagre_tab_remove_from_notebook (tab);
 	      vinagre_utils_show_error (_("Authentication error"),
 					_("A password is required in order to access this machine."),
-					GTK_WINDOW (window));
+					window);
 	      goto out;
 	    }
 	}
 
+      vinagre_tab_set_save_credentials (tab, save_in_keyring);
+
 out:
       g_free (username);
       g_free (password);
+      g_free (host);
     }
 }
 
@@ -690,6 +608,20 @@ create_connected_actions (VinagreVncTab *tab)
   list = g_slist_append (list, a);
   tab->priv->scaling_action = a->action;
 
+  /* View->Keep Ratio */
+  a = g_new (VinagreTabUiAction, 1);
+  a->paths = g_new (gchar *, 2);
+  a->paths[0] = g_strdup ("/MenuBar/ViewMenu");
+  a->paths[1] = NULL;
+  a->action = GTK_ACTION (gtk_toggle_action_new ("VNCViewKeepRatio",
+						 _("_Keep Aspect Ratio"),
+						 _("Keeps the screen aspect ratio when using scaling"),
+						 NULL));
+  gtk_action_set_sensitive (a->action, FALSE);
+  g_signal_connect (a->action, "activate", G_CALLBACK (view_keep_ratio_cb), tab);
+  list = g_slist_append (list, a);
+  tab->priv->keep_ratio_action = a->action;
+
   /* View->View Only */
   a = g_new (VinagreTabUiAction, 1);
   a->paths = g_new (gchar *, 3);
@@ -848,52 +780,10 @@ vnc_tab_clipboard_cb (GtkClipboard *cb, GdkEvent *event, VinagreVncTab *vnc_tab)
   g_free (text);
 }
 
-/*
- * Called when the main container widget's size has been set.
- * It attempts to fit the VNC widget into this space while
- * maintaining aspect ratio
- *
- * Code borrowed from from virt-viewer, thanks Daniel Berrange :)
- */
-static void
-vnc_tab_resize_align (GtkWidget *widget,
-		      GtkAllocation *alloc,
-		      VinagreVncTab *vnc_tab)
-{
-  double desktopAspect = (double)vnc_display_get_width (VNC_DISPLAY (vnc_tab->priv->vnc)) / (double)vnc_display_get_height (VNC_DISPLAY (vnc_tab->priv->vnc));
-  double scrollAspect = (double)alloc->width / (double)alloc->height;
-  int height, width;
-  GtkAllocation child;
-  int dx = 0, dy = 0;
-
-  if (!vnc_display_is_open (VNC_DISPLAY (vnc_tab->priv->vnc)))
-    return;
-
-  if (scrollAspect > desktopAspect)
-    {
-      width = alloc->height * desktopAspect;
-      dx = (alloc->width - width) / 2;
-      height = alloc->height;
-    }
-  else
-    {
-      width = alloc->width;
-      height = alloc->width / desktopAspect;
-      dy = (alloc->height - height) / 2;
-    }
-
-  child.x = alloc->x + dx;
-  child.y = alloc->y + dy;
-  child.width = width;
-  child.height = height;
-  gtk_widget_size_allocate(vnc_tab->priv->vnc, &child);
-}
-
 static void
 vinagre_vnc_tab_init (VinagreVncTab *vnc_tab)
 {
   GtkClipboard *cb;
-  GtkWidget *align;
 
   vnc_tab->priv = VINAGRE_VNC_TAB_GET_PRIVATE (vnc_tab);
   vnc_tab->priv->clipboard_str = NULL;
@@ -902,13 +792,11 @@ vinagre_vnc_tab_init (VinagreVncTab *vnc_tab)
 
   /* Create the vnc widget */
   vnc_tab->priv->vnc = vnc_display_new ();
-  align = gtk_alignment_new (0.5, 0.5, 1, 1);
+  vnc_tab->priv->align = gtk_alignment_new (0.5, 0.5, 1, 1);
+  vnc_tab->priv->signal_align = 0;
+  gtk_container_add (GTK_CONTAINER (vnc_tab->priv->align), vnc_tab->priv->vnc);
 
-  g_signal_connect(align, "size-allocate",
-		  G_CALLBACK (vnc_tab_resize_align), vnc_tab);
-  gtk_container_add (GTK_CONTAINER (align), vnc_tab->priv->vnc);
-
-  vinagre_tab_add_view (VINAGRE_TAB (vnc_tab), align);
+  vinagre_tab_add_view (VINAGRE_TAB (vnc_tab), vnc_tab->priv->align);
 
   g_signal_connect (vnc_tab->priv->vnc,
 		    "vnc-connected",
@@ -1035,6 +923,7 @@ vinagre_vnc_tab_set_scaling (VinagreVncTab *tab, gboolean active) {
 
   gtk_toggle_tool_button_set_active (GTK_TOGGLE_TOOL_BUTTON (tab->priv->scaling_button),
 				     active);
+  gtk_action_set_sensitive (tab->priv->keep_ratio_action, active);
 
   if (active)
     gtk_widget_set_size_request (tab->priv->vnc, 0, 0);
@@ -1060,7 +949,6 @@ vinagre_vnc_tab_set_viewonly (VinagreVncTab *tab, gboolean active) {
   vnc_display_set_read_only (VNC_DISPLAY (tab->priv->vnc), active);
   gtk_toggle_tool_button_set_active (GTK_TOGGLE_TOOL_BUTTON (tab->priv->viewonly_button),
 				     active);
-
 }
 
 gboolean
@@ -1070,6 +958,75 @@ vinagre_vnc_tab_get_viewonly (VinagreVncTab *tab) {
   return vnc_display_get_read_only (VNC_DISPLAY (tab->priv->vnc));
 }
 
+/*
+ * Called when the main container widget's size has been set.
+ * It attempts to fit the VNC widget into this space while
+ * maintaining aspect ratio
+ *
+ * Code borrowed from from virt-viewer, thanks Daniel Berrange :)
+ */
+
+static void
+vnc_tab_resize_align (GtkWidget *widget,
+		      GtkAllocation *alloc,
+		      VinagreVncTab *vnc_tab)
+{
+  double desktopAspect = (double)vnc_display_get_width (VNC_DISPLAY (vnc_tab->priv->vnc)) / (double)vnc_display_get_height (VNC_DISPLAY (vnc_tab->priv->vnc));
+  double scrollAspect = (double)alloc->width / (double)alloc->height;
+  int height, width;
+  GtkAllocation child;
+  int dx = 0, dy = 0;
+
+  if (!vnc_display_is_open (VNC_DISPLAY (vnc_tab->priv->vnc)))
+    return;
+
+  if (scrollAspect > desktopAspect)
+    {
+      width = alloc->height * desktopAspect;
+      dx = (alloc->width - width) / 2;
+      height = alloc->height;
+    }
+  else
+    {
+      width = alloc->width;
+      height = alloc->width / desktopAspect;
+      dy = (alloc->height - height) / 2;
+    }
+
+  child.x = alloc->x + dx;
+  child.y = alloc->y + dy;
+  child.width = width;
+  child.height = height;
+  gtk_widget_size_allocate(vnc_tab->priv->vnc, &child);
+}
+
+void
+vinagre_vnc_tab_set_keep_ratio (VinagreVncTab *tab, gboolean active)
+{
+  g_return_if_fail (VINAGRE_IS_VNC_TAB (tab));
+
+  if (tab->priv->signal_align > 0)
+    g_signal_handler_disconnect (tab->priv->align, tab->priv->signal_align);
+
+  if (active)
+    tab->priv->signal_align = g_signal_connect (tab->priv->align,
+						"size-allocate",
+						G_CALLBACK (vnc_tab_resize_align),
+						tab);
+  else
+    tab->priv->signal_align = 0;
+
+  gtk_widget_queue_resize_no_redraw (tab->priv->align);
+}
+
+gboolean
+vinagre_vnc_tab_get_keep_ratio (VinagreVncTab *tab)
+{
+  g_return_val_if_fail (VINAGRE_IS_VNC_TAB (tab), FALSE);
+
+  return vinagre_vnc_connection_get_keep_ratio (VINAGRE_VNC_CONNECTION (vinagre_tab_get_conn (VINAGRE_TAB (tab))));
+}
+
 gboolean
 vinagre_vnc_tab_is_pointer_grab (VinagreVncTab *tab)
 {
